/* 
 *           DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *                    Version 2, December 2004
 *  
 * Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>
 * 
 * Everyone is permitted to copy and distribute verbatim or modified
 * copies of this license document, and changing it is allowed as long
 * as the name is changed.
 *  
 *            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 *  
 *  0. You just DO WHAT THE FUCK YOU WANT TO.
 */
 
/*
 * Hand crafted IPS patch generation.
 * author : MooZ
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#include "ips.h"

#define FONT_ELEMENCOUNT_OFFSET 0x2ebe7

#define FONT_DATA_OFFSET 0x2ebeb
#define FONT_DATA_SIZE 1208

#define FONT_ENCODING_TABLE_OFFSET 0x2f2a3
#define FONT_ENCODING_TABLE_SIZE 0x10

#define JUMP_TABLE_OFFSET 0x11814
#define JUMP_TABLE_START 0x564C
#define JUMP_TABLE_ELEMENT_COUNT 27

#define STRING_START_OFFSET 0x1184C
#define STRING_END_OFFSET 0x11d61

#define FREE_SPACE_START_OFFSET 0x1e60
#define FREE_SPACE_END_OFFSET 0x21df
#define FREE_SPACE_SIZE 899

#define CODE_START_OFFSET 0x4be2
#define CODE_END_OFFSET 0x4cb4
#define CODE_SIZE 211

#define LEVEL_NAMES_DATA_OFFSET 0x3c482
#define LEVEL_NAMES_DATA_SIZE 504

#define LEVEL_NAMES_ENCODING_TABLE_OFFSET 0x3d4ab
#define LEVEL_NAMES_ENCODING_TABLE_SIZE 21

#define LEVEL_NAMES_PTR_OFFSET 0x5183
#define LEVEL_NAMES_PTR_SIZE 18

#define LEVEL_NAMES_LIST_OFFSET 0x5195
#define LEVEL_NAMES_LIST_SIZE 51

#define TEXT_BOX_BAT_POS_OFFSET 0x11D6B
#define TEXT_BOX_BAT_POS_SIZE   8

#define CREDITS_OFFSET   0x11FFE
#define CREDITS_SIZE     0xE2

uint8_t g_jumpTable[JUMP_TABLE_ELEMENT_COUNT*2];

uint8_t* g_text;
size_t g_textLen;

size_t g_pairCount;

void InitText ()
{
	g_text = NULL;
	g_textLen = 0;

	g_pairCount = 0;
}

void ClearText ()
{
	if( g_text != NULL )
	{
		free( g_text );
		g_text = NULL;
	}
	
	g_textLen = 0;
}

int8_t LoadText ()
{
	int8_t err = IPS_OK;
	size_t nRead;
	FILE *input = fopen( "text.bin", "rb" );
	if( input == NULL )
	{
		perror( strerror( errno ) );
		return IPS_ERROR;
	}
	
	fseek( input, 0, SEEK_END );
	g_textLen  = ftell( input );
	fseek( input, 0, SEEK_SET );
	g_textLen -= ftell( input );
	
	g_text = (uint8_t*)malloc( g_textLen );
	if( g_text == NULL )
	{
		perror( strerror( errno ) );
		err = IPS_ERROR;
	}
	else
	{
		nRead = fread( g_text, 1, g_textLen, input );
		if( nRead != g_textLen )
		{
			perror( strerror( errno ) );
			free( g_text );
			g_text = NULL;
			err = IPS_ERROR;
		}
	}
	
	fclose( input );

	return err;
}

int8_t AddDataToIPS( FILE *out, const char *inFilename, int inOffset, int inSize )
{
	int8_t err;
	
	FILE *input;
	uint8_t *data;
	size_t size, nRead;
	
	input = fopen( inFilename, "rb");
	if( input == NULL )
	{
		perror( strerror( errno ) );
		return IPS_ERROR;
	}
	
	fseek( input, 0, SEEK_END );
	size  = ftell( input );
	fseek( input, 0, SEEK_SET );
	size -= ftell( input );
	/*
	if( size != inSize )
	{
		fprintf( stderr, "File is too large (expected %d, got %d)\n", inSize, size );
		return IPS_ERROR;
	}*/
	
	data = (uint8_t*)malloc( size );
	if( data == NULL )
	{
		perror( strerror( errno ) );
		return IPS_ERROR;
	}
	
	nRead = fread( data, 1, size, input );
	if( nRead != size )
	{
		perror( strerror( errno ) );
		free( data );
		return IPS_ERROR;
	}
	
	err = IPSWriteRecord ( out, inOffset, inSize, data );

	free( data );
	return err;
}

int8_t AddTablesAndCodeToIPS( FILE *out )
{
	uint8_t *pairs;
	uint8_t code[CODE_SIZE];
	int8_t err;
	
	FILE *input;

	input = fopen( "text.asm", "rb" );
	if( input == NULL )
	{
		perror( strerror( errno ) );
		return IPS_ERROR;
	}
	
	fscanf( input, "BPE_PAIR_COUNT=%d\n", &g_pairCount );
	g_pairCount *= 2;
	fclose( input );

	pairs = (uint8_t*)malloc( g_pairCount );
	if( pairs == NULL )
	{
		perror( strerror( errno ) );
		return IPS_ERROR;
	}	

	input = fopen( "font_bpe.pce", "rb" );
	if( input == NULL )
	{
		perror( strerror( errno ) );
		free( pairs );
		return IPS_ERROR;
	}
	
	fseek( input, FREE_SPACE_START_OFFSET, SEEK_SET );
	fread( pairs, 1, g_pairCount, input );
	
	fseek( input, CODE_START_OFFSET, SEEK_SET );
	fread( code, 1, CODE_SIZE, input );
	
	fclose( input );
	
	err = IPSWriteRecord ( out, FREE_SPACE_START_OFFSET, g_pairCount, pairs );
	free( pairs );

	if( err != IPS_OK )
	{
		return err;
	}
	
	return IPSWriteRecord ( out, CODE_START_OFFSET, CODE_SIZE, code );
}

int8_t AddTextToIPS( FILE *out )
{
	size_t totalSize = FONT_DATA_SIZE + FREE_SPACE_SIZE - g_pairCount;
	size_t first = 0;
	int8_t err = IPS_OK;
	
	if( g_textLen > totalSize )
	{
		fprintf( stderr, "Text is too big\n" );
	}
	
	if( g_textLen < FONT_DATA_SIZE )
	{
		return IPSWriteRecord ( out, STRING_START_OFFSET, g_textLen, g_text );
	}

	for(first = FONT_DATA_SIZE; (first > 0) && (g_text[first] != 0xff); --first);
	
	if( first < 1 )
	{
		fprintf( stderr, "String is far too long!\n" );
		return IPS_ERROR;
	}
	if( (g_textLen - first) > (FREE_SPACE_SIZE - g_pairCount) )
	{
		fprintf( stderr, "Not enough free space for text!\n" );
		return IPS_ERROR;
	}
	
	err = IPSWriteRecord ( out, FREE_SPACE_START_OFFSET + g_pairCount, g_textLen - first, g_text + first );
	
	if( err != IPS_OK )
	{
		return err;
	}
	else
	{
		return IPSWriteRecord ( out, STRING_START_OFFSET, first, g_text );
	}
}

int8_t AddJumpTableToIPS( FILE *out )
{
	int i;
	size_t j = 0;

	if( g_textLen < FONT_DATA_SIZE )
	{
		for(i=0, j=0; i<JUMP_TABLE_ELEMENT_COUNT; ++i)
		{
			g_jumpTable[(2*i)  ] = ((j + JUMP_TABLE_START)     ) & 0xff;
			g_jumpTable[(2*i)+1] = ((j + JUMP_TABLE_START) >> 8) & 0xff;
			
			for(; (j < g_textLen) && (g_text[j] != 0xff); ++j )
			{}
			
			++j;
		}
	}
	else
	{
		size_t first, start;
		start = FREE_SPACE_START_OFFSET + g_pairCount;
		
		for(first = FONT_DATA_SIZE; (first > 0) && (g_text[first] != 0xff); --first);
		
		for(i=0, j=0; i<JUMP_TABLE_ELEMENT_COUNT; ++i)
		{
			g_jumpTable[(2*i)  ] = ((j + JUMP_TABLE_START)     ) & 0xff;
			g_jumpTable[(2*i)+1] = ((j + JUMP_TABLE_START) >> 8) & 0xff;
			
			for(; (j < first) && (g_text[j] != 0xff); ++j )
			{}
			
			++j;
			if(j >= first) break;
		}
		
		for(; i<JUMP_TABLE_ELEMENT_COUNT; ++i)
		{
			g_jumpTable[(2*i)  ] = ((j + start)     ) & 0xff;
			g_jumpTable[(2*i)+1] = ((j + start) >> 8) & 0xff;
			
			for(; (j < g_textLen) && (g_text[j] != 0xff); ++j )
			{}
			++j;
		}
	}
	
	return IPSWriteRecord ( out, JUMP_TABLE_OFFSET, JUMP_TABLE_ELEMENT_COUNT*2, g_jumpTable );
}

int main(int argc, char **argv)
{
	int i, err = 0;
	FILE *ipsOutput;
	
	const struct
	{
		char *filename;
		int offset;
		int size;
	} data[] = 
	{
		{ "font.compress.bin", FONT_DATA_OFFSET, FONT_DATA_SIZE },
		{ "font.compress.bin.hed", FONT_ENCODING_TABLE_OFFSET, FONT_ENCODING_TABLE_SIZE },
		{ "levels.compress.bin", LEVEL_NAMES_DATA_OFFSET, LEVEL_NAMES_DATA_SIZE },
		{ "levels.compress.bin.hed", LEVEL_NAMES_ENCODING_TABLE_OFFSET, LEVEL_NAMES_ENCODING_TABLE_SIZE },
		{ "levels.ptr.bin", LEVEL_NAMES_PTR_OFFSET, LEVEL_NAMES_PTR_SIZE },
		{ "levels.index.bin", LEVEL_NAMES_LIST_OFFSET, LEVEL_NAMES_LIST_SIZE},
		{ "text.bat.bin", TEXT_BOX_BAT_POS_OFFSET, TEXT_BOX_BAT_POS_SIZE},
		{ "credits.bin", CREDITS_OFFSET, CREDITS_SIZE }
	};
	const int dataCount = sizeof(data)/sizeof(data[0]);
	
	InitText ();
	
	if(argc != 2)
	{
		fprintf( stderr, "%s out.ips\n", argv[0] );
		return 1;
	}

	if( LoadText() != IPS_OK )
	{
		return 1;
	}

	if( IPSWriteBegin( &ipsOutput, argv[1] ) != IPS_OK )
	{
		fprintf( stderr, "Failed to open %s\n", argv[1] );
		return 1;
	}

	if( AddTablesAndCodeToIPS( ipsOutput ) != IPS_OK )
	{
		fprintf( stderr,"Add tables and code failed!\n");
		err = 1;
		goto error;
	}
	
	if( AddTextToIPS( ipsOutput ) != IPS_OK )
	{
		fprintf( stderr,"Add text failed!\n");
		err = 1;
		goto error;
	}

	if( AddJumpTableToIPS( ipsOutput ) != IPS_OK )
	{
		fprintf( stderr, "Add jump table failed!\n");
		err = 1;
		goto error;
	}

	for(i=0; i<dataCount; ++i)
	{
		if( AddDataToIPS( ipsOutput, data[i].filename, data[i].offset, data[i].size ) != IPS_OK )
		{
			fprintf( stderr, "Add data %d failed\n", i);
			err = 1;
			goto error;
		}
	}
	
	IPSWriteEnd( &ipsOutput );
	
error:	
	ClearText();
	
	return err;
}
