/* 
 *           DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *                    Version 2, December 2004
 *  
 * Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>
 * 
 * Everyone is permitted to copy and distribute verbatim or modified
 * copies of this license document, and changing it is allowed as long
 * as the name is changed.
 *  
 *            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 *  
 *  0. You just DO WHAT THE FUCK YOU WANT TO.
 */

/*
 * Marchen Maze data decoder (used for tiles and sprites)
 * author : MooZ
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>

#define RAW 0
#define RLE 2
#define XOR 3

int decodeRLE( FILE* in, uint8_t buffer[32] )
{
	uint8_t header[4];
	int i, j, k;
	
	fread( header, 1, 4, in );
	
	for( i=0, k=0; i<4; ++i )
	{
		for( j=0; j<8; ++j, ++k )
		{
			if( header[i] & 0x01 )
			{
				fread( buffer+k, 1, 1, in );
			}
			else
			{
				buffer[k] = 0;
			}
			header[i] >>= 1;
		}
	} 
	
	return 1;
}

int decodeXOR( FILE *in, uint8_t buffer[32] )
{
	int err;
	int x, y;
	
	err = decodeRLE( in, buffer );
	if( !err ) return err;
	
	for(y=7, x=0; y!=0; --y)
	{
		buffer[x+0x02] ^= buffer[x];
		buffer[x+0x03] ^= buffer[x+0x01];

		buffer[x+0x12] ^= buffer[x+0x10];
		buffer[x+0x13] ^= buffer[x+0x11];
		x+=2;
	}
	
	return 1;
}

int main( int argc, char** argv )
{
	int err;
	FILE *in, *out, *header;
	uint8_t buffer[32];
	
	if( argc != 4 )
	{
		fprintf( stderr, "%s data header out\n", argv[0] );
		return 1;
	}
	
	in = fopen( argv[1], "rb" );
	if( in == NULL )
	{
		fprintf( stderr, "Failed to open %s : %s\n", argv[1], strerror( errno ) );
		return 1;
	}

	header = fopen( argv[2], "rb" );
	if( header == NULL )
	{
		fprintf( stderr, "Failed to open %s : %s\n", argv[2], strerror( errno ) );
		fclose(in);
		return 1;
	}

	err = decodeXOR( in, buffer );


	out = fopen( argv[3], "wb" );
	if( out == NULL )
	{
		fprintf( stderr, "Failed to open %s : %s\n", argv[3], strerror( errno ) );
		fclose(in);
		fclose(header);
		return 1;
	}
	
	fwrite( buffer, 1, 32, out );
	
	fclose(in);
	fclose(header);
	fclose(out);
			
	return 0;
}
