/* 
 *           DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *                    Version 2, December 2004
 *  
 * Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>
 * 
 * Everyone is permitted to copy and distribute verbatim or modified
 * copies of this license document, and changing it is allowed as long
 * as the name is changed.
 *  
 *            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 *  
 *  0. You just DO WHAT THE FUCK YOU WANT TO.
 */

/*
 * Encode script (bpe etc...)
 * author : MooZ
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#include "bpe.h"

const struct {
	char *str;
	unsigned char code; 
} g_tokens[4] = {
	{ "FE-CR", 0xfe },
	{ "FC-CR", 0xfc },
	{ "FD-CR", 0xfd },
	{ "END",   0xff },
};

const size_t g_dataStringSize = 0x10000;

#define IGNORE_LEN 2
#define ENCODED_CHAR_MIN 103
#define ENCODED_CHAR_MAX 250
#define DICTIONARY_LEN (ENCODED_CHAR_MAX-ENCODED_CHAR_MIN)

#define EVIL_HARDCODED_COUNT 27

const uint8_t g_ignore[IGNORE_LEN] = { 0xfc, 0xff };
uint8_t g_dictionary[DICTIONARY_LEN];

struct CodeBuffer
{
	unsigned char *data;
	int len;
	int last;
};

int init( struct CodeBuffer *buff, int len )
{
	buff->data = (unsigned char*)malloc( len );
	if(buff->data == NULL)
	{
			perror( strerror( errno ) );
			return 0;
	}
	
	buff->len  = len;
	buff->last = 0;
	
	return len;
}

void reset( struct CodeBuffer* buff )
{
	buff->last = 0;
}

void clear( struct CodeBuffer *buff )
{
	buff->len = buff->last = 0;
	if(buff->data != NULL)
	{
		free(buff->data);
		buff->data = NULL;
	}
}

int push( struct CodeBuffer *buff, unsigned char code )
{
	if(buff->last == buff->len)
	{
		unsigned char* tmp = (unsigned char*)realloc(buff->data, 2*buff->len);
		if(tmp == NULL)
		{
			perror( strerror( errno ) );
			return 0;
		}
		
		buff->len *= 2;
		buff->data = tmp;
	}
	
	buff->data[buff->last++] = code;
	
	return 1;
}

int translate( char *text, int len, struct CodeBuffer *buff )
{
	int i, j, tokenLen;
	char *ptr;

	unsigned char code = 0x00;
	
	for(i=0, ptr=text; (i<len) && (code != 0xff); ++i, ++ptr)
	{
		switch(*ptr)
		{
			case '[':
				for( j=0; j<4; ++j )
				{
					tokenLen = strlen( g_tokens[j].str );
					if( strncmp( g_tokens[j].str, ptr+1, tokenLen ) == 0 )
					{
						if( *(ptr+1+tokenLen) == ']' )
						{
							ptr += tokenLen+1;
							i += tokenLen+1;
							code = g_tokens[j].code;	
							break; 
						}
					}
				}
				if(j == 4)
				{
					code = *ptr - '[' + 49;
				}
			break;
			
			case '\r':
			case '\n':
				continue;
			break;
						
			default:
				if( (*ptr >= ' ') && (*ptr <= '/') )
				{
					code = *ptr - ' ';
				}
				else if( (*ptr >= '0') && (*ptr <= '9') )
				{
					code = *ptr - '0' + 64;
				}
				else if( (*ptr >= ':') && (*ptr <= '@') )
				{
					code = *ptr - ':' + 16;
				}
				else if( (*ptr >= 'A') && (*ptr <= 'Z') )
				{
					code = *ptr - 'A' + 74;
				}
				else if( (*ptr >= '[') && (*ptr <= '_') )
				{
					code = *ptr - '[' + 49;
				}
				else if( (*ptr >= 'a') && (*ptr <= 'z') )
				{
					code = *ptr - 'a' + 23;
				}
				else if( (*ptr >= '{') && (*ptr <= '}') )
				{
					code = *ptr - '{' + 54;
				}
				else if( *ptr == 0xA9 )
				{
					code = 102;
				}
				else
				{
					code = 0x00;
				}
				break;
		}

		if( !push(buff, code) )
		{
			return 0;
		}
	}
	return i;
}

int initCompressor( struct BPEData* bpe )
{
	uint8_t i;
	
	for(i=0; i<DICTIONARY_LEN; ++i)
	{
		g_dictionary[i] = ENCODED_CHAR_MIN+i;
	}
	
	if(bpeInit( bpe, g_dictionary, DICTIONARY_LEN ) != BPE_OK)
	{
		return 0;
	}
	
	return 1;
}

int compress( struct CodeBuffer* buff, struct BPEData* bpe )
{
	if( bpeEncode( (char*)buff->data, buff->last, g_ignore, IGNORE_LEN, bpe) != BPE_OK )
	{
		return 0;
	}

	return 1;
}

int checkCompressedData( struct CodeBuffer* buff, struct BPEData* bpe )
{
	unsigned int i, j, k, top;
	uint8_t stack[64], c;

	k = 0;
	for(i=0; i<bpe->compressedStringLength; ++i)
	{
		c = bpe->compressedString[i];
		top = 0;
		stack[top++] = c;

		while(top>0)
		{
			c = stack[--top];
			for(j=0; (j<DICTIONARY_LEN) && (c!=g_dictionary[j]); ++j);
			if(j<DICTIONARY_LEN)
			{
				stack[top++] = bpe->pairs[j].second;
				stack[top++] = bpe->pairs[j].first;
			}
			else
			{
				if(k >= buff->last)
				{
					fprintf(stderr,"ARG! Generated too much data! %d %d\n", k, buff->last );
					return 0;
				}
				if(c != buff->data[k])
				{
					fprintf(stderr,"ARG! Compression check failed! %c %c\n", c ,buff->data[k]);
					return 0;
				}
				++k;
			}
		}
	}
	
	if(k < buff->last)
	{
		fprintf(stderr,"ARG! Compression check failed!\n");
		return 0;
	}		
	
	return 1;
}

int main(int argc, char **argv)
{
	struct CodeBuffer buff;
	int nTranslated, remaining, i, err;

	FILE *script, *out;
	char *scriptText, *ptr;
	size_t scriptTextLen;
	
	struct BPEData bpe;
	 
	if(argc != 2)
	{
		fprintf( stderr, "%s script\n", argv[0] );
		return 1;
	}

	if( !initCompressor( &bpe ) )
	{
		return 1;
	}

	script = fopen( argv[1], "rb" );
	if( script == NULL )
	{
		perror( strerror( errno ) );
		return 1;
	}

	fseek( script, 0, SEEK_END );
	scriptTextLen = ftell( script );
	fseek( script, 0, SEEK_SET );
	scriptTextLen -= ftell( script );

	scriptText = (char*)malloc( scriptTextLen );
	if( scriptText == NULL )
	{
		perror( strerror( errno ) );
		fclose( script );
		return 1;
	}

	fread( scriptText, 1, scriptTextLen, script );

	fclose( script );
		
	init(&buff, g_dataStringSize);

	remaining = scriptTextLen;
	ptr = scriptText;
	i = 0;
	do
	{			
		nTranslated = translate( ptr, remaining, &buff );
		remaining -= nTranslated;
		ptr += nTranslated;
		++i;
	}while( nTranslated && (remaining >= 0) && (i<EVIL_HARDCODED_COUNT));

	err = 0;
	if( !compress( &buff, &bpe ) )
	{
		err = 1;
		fprintf(stderr, "ARG\n");
	}

	if( !checkCompressedData( &buff, &bpe ) )
	{
		err = 1;
		fprintf(stderr, "Arg check failed!\n");
	}
	else
	{
		out = fopen( "text.asm", "wb" );
		fprintf( out,
		         "BPE_PAIR_COUNT=%d\n"
		         "BPE_COMPRESSED_STRING_LEN=%d\n"
		         "BPE_UNCOMPRESSED_STRING_LEN=%d\n"
		         "ENCODED_CHAR_MIN=%d\n"
		         "ENCODED_CHAR_MAX=%d\n",
		         bpe.pairCount,
		         bpe.compressedStringLength, buff.last,
		         ENCODED_CHAR_MIN, ENCODED_CHAR_MAX);
			
		fprintf( out, "_bpe_first:\n\t.db " );
		for(i=0; i<bpe.pairCount; ++i)
		{
			fprintf( out, "$%02x", bpe.pairs[i].first );
			if((i & 7) == 7) fprintf( out, "\n\t.db ");
			else if( i<(bpe.pairCount-1) ) fprintf( out, ", " );
		}
		fprintf( out, "\n_bpe_second:\n\t.db " );
		for(i=0; i<bpe.pairCount; ++i)
		{
			fprintf( out, "$%02x", bpe.pairs[i].second );
			if((i & 7) == 7) fprintf( out, "\n\t.db " );
			else if( i<(bpe.pairCount-1) ) fprintf( out, ", " );
		}
		fprintf( out, "\n" );
		fclose( out );
		
		out = fopen( "text.bin", "wb" );
		fwrite( bpe.compressedString, 1, bpe.compressedStringLength, out );
		fclose( out );
	}
	
	bpeClear( &bpe );	
	clear(&buff);
	
	return err;
}
